/*  zmneanred -- Reduce image size by averaging blocks of pixels

NAME:
    zmneanred

PURPOSE:
    Called by ZOOM for subcommand MEANRED to reduce an image by
    averaging pixels within rectangular blocks.

ARGUMENTS:
(&vblock,hosin[0],hosout,nmrimg,window,bands,nbands,
		totbnd,odtype,xfact,yfact,msgtim,minmax,nol,nos);
    struct PARBLK
	    *vblock	(in)	TAE parameter block.
    char hosin[][CMLEN]	(in)	Host names of input images.
    char hosout[][CMLEN](in)	Host name of output image.
    long    nmrimg	(in)    Number of input images.
    long    (*window)[4] (in)	Window specification output image.
    long    (*bands)[MAXBND+1]
			(in)	Bands for each input image.
    long    *nbands	(in)	Number of bands per input image.
    long    totbnd	(in)	Total number of input bands.
    long    odtype	(in)	Data type of output image.
    long    xfact	(in)	Expansion factor for the x-axis.
    long    yfact	(in)	Expansion factor for the y-axis.
    long    msgtim	(in)	Processing progress message interval.
    char    *minmax	(in)	Min/max flag.

RETURN VALUE:
    E_SUCC	Normal completion.
		(Function aborts on fatal error).

FILE REFERENCES:
    hosin	Input image file(s).
    hosout	Output image file.

PROGRAM HISTORY:                                  
    1999 Apr	Initial version, R. A. White, PSU/ESSC.
    2000 Jun	Fixed computation of upper left corner, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.6.

PROJECT:
    LAS at PSU/ESSC

INFORMATIONAL AND ERROR MESSAGES:

    These are described in zoom.pdf.

ALGORITHM DESCRIPTION:
    Open input image(s) for sequential access
    Open output image for sequential access
    Allocate buffers
    Open input image group
    Open output image group
    Start processing messages
    Initialize min/max buffers
    For each input line
        For each input band
	    For each input pixel
	        If first input pixel and line in block for output pixel
		    Copy input value to output
		Else
		    Add input value to output
	If last input line contributing to current output line
	    For each input band
	        For each output pixel
	            Divide sum of input values by size of block
		    If floating point value is to be rounded to integer
		        Round fraction >= 0.5 upwards
		Update output min/max
	Write output line
    Stop processing messages
    Close image files
    Update DDR with new scale and corner coordinates
    Insert band minima and maxima
    									
***************   End of Prolog for zmneanred   ***************	*/
									/*
Header Files:			       					*/

#include	<ctype.h>	/* Character classification functions	*/
#include	<math.h>	/* Mathematical functions		*/
#include	<stdio.h>	/* Standard I/O definitions		*/
#include	<stdlib.h>	/* Standard C functions			*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"util.h"	/* LAS utility functions		*/
#include	"pixman.h"	/* LAS pixel-manipulation functions	*/
						   			/*
Local Program Constants and Macros:					*/
#define  ABORT \
	   c_lasmsg (TRUE, "zoom-fatal", "Fatal error encountered")
									/*
External Variable Declarations:						*/
    extern long	curent;		/* current processing line */
									/*
***************   Start of Function zmneanred ()   ***************	*/

long FUNCTION zmmeanred (
    struct PARBLK *vblock,	/* TAE parameter block */
    char hosin[][CMLEN],	/* Host names of input images */
    char hosout[][CMLEN],	/* Host name of output image */
    long	nmrimg,		/* Number of input images */
    long	(*window)[4],	/* Window specification output image */
    long	(*bands)[MAXBND+1], /* Bands for each input image */
    long	*nbands,	/* Number of bands per input image */
    long	totbnd,		/* Total number of input bands */
    long	odtype,		/* Data type of output image */
    long	xfact,		/* Expansion factor for the x-axis */
    long	yfact,		/* Expansion factor for the y-axis */
    long	msgtim,		/* Processing progress message interval */
    char	*minmax)	/* Min/max flag */
    {
	                                              			/*
Local Variable Definitions:						*/

    long	access;		/* Code for image access type */
    long	band;		/* Sequential input band number */
    struct BDDR *bddr;		/* Temporary storage for BDR data */
    long	blksize;	/* Size of output block for input pixel */
    struct DDR	ddr;		/* Newly created output ddr */
    long	dtype;		/* Code for internal data type */
    struct FDESC *fdi[MAXIMG+1]; /* File descriptor(s) for input image */
    struct FDESC *fdo[2];	/* File descriptor for output image */
    struct GDESC *gdi;		/* Group descriptor for input image(s) */
    struct GDESC *gdo;		/* Group descriptor for output image */
    long	ibufsize;	/* Size of input buffer, in bytes */
    long	image;		/* Current input image */
    float	 *inbuf;	/* Buffer for input image data */
    long	is;		/* Input sample number */
    long	iline;		/* Current input line */
    double	*mmval[2];	/* Arrays for output minima and maxima */
    long	msgformat;	/* Processing progress message format */
    long	nli;		/* Number of lines in input window */
    long	nlo;		/* Number of output lines */
    long	nsi;		/* Number of samples per input line */
    long	nso;		/* Number of samples per output line */
    long	obufsize;	/* Size of output buffer, in bytes */
    long	offseti[MAXBND];/* Offsets into input buffer */
    long	offseto[MAXBND];/* Offsets into output buffer */
    long	one = 1;	/* Constant one */
    long	opt[2];		/* Options for c_eopen () functions */
    long	os;		/* Output sample number */
    float	*outbuf;	/* Buffer for output data */
    long	roundflag;	/* Flag for integer output data type */
    double	xratio;		/* Ratio of output to input x */
    double	yratio;		/*  and y pixel dimensions    */
    long	zero = 0;	/* Constant zero */

/*  Open input image(s) for sequential access  */

    access = IREAD;
    opt[0] = opt[1] = 0;
    nli = window[0][2];
    nsi = window[0][3];
    nlo = nli / yfact;
    nso = nsi / xfact;
    nli = nlo * yfact;
    nsi = nso * xfact;
    dtype = EREAL;

    for (band = 0, image = 0; image < nmrimg; image++)
	{
	if (c_eopens (&fdi[image], hosin[image], bands[image],
	  &window[image][0], &window[image][1] , &nli ,&nsi,
	  &offseti[band], &access, &dtype, &zero, opt) != E_SUCC)
	    ABORT;
	band += nbands[image];
	}
    fdi[nmrimg] = 0;

/*  Open output image for sequential access  */

    access = IWRITE;
    opt[0] = odtype;
    if (c_eopens (&fdo[0], hosout[0], &totbnd, &one, &one, &nlo, &nso,
     offseto, &access, &dtype, &zero, opt) != E_SUCC)
	    ABORT;
    fdo[1] = 0;

/*  Allocate buffers  */

    if ((inbuf = (float *) c_ealloc (fdi, &ibufsize)) == NULL)
	c_lasmsg (TRUE, "zoom-alloc",
	  "Error allocating dynamic memory");
    if ((outbuf = (float *) c_ealloc (fdo, &obufsize)) == NULL)
	c_lasmsg (TRUE, "zoom-alloc",
	  "Error allocating dynamic memory");

/*  Define input and output image groups  */

    if (c_egroup (fdi, &gdi, (unsigned char*) inbuf, &ibufsize)
     != E_SUCC)
	ABORT;
    if (c_egroup (fdo, &gdo, (unsigned char*) outbuf, &obufsize)
     != E_SUCC)
	ABORT;

/*  Initialize output min/max array  */

    if ((mmval[0] = (double *) malloc (2 * totbnd * sizeof (double)))
      == NULL)
	c_lasmsg (TRUE, "zoom-alloc",
	  "Error allocating dynamic memory");
    mmval[1] = mmval[0] + totbnd;
    if (c_inmnmx (&odtype, &totbnd, mmval[0], mmval[1]) != E_SUCC)
	ABORT;

/*  Start processing messages  */

    if (msgtim > 0)
	{
	msgformat = (totbnd > 1) ? ALLCAT : GEN2;
	c_inittm ("ZOOM", "processing", "input line", "band", "", &nli,
	  &totbnd, &msgtim, &msgformat);
	curent = 1;
	c_startm ();
	}

/*  Process successive bands for one line at a time  */

    roundflag = ((odtype == EREAL) || (odtype == EDOUBLE)) ? 0 : 1;
    if (c_estep (&gdo) != E_SUCC)
	ABORT;
    for (iline = 0; iline < nli; iline++, curent++)
	{
	if (c_estep (&gdi) != E_SUCC)
	    {
	    c_egclse(&gdo);
	    c_clean (vblock, "zoom", hosout, &one);
	    }
	for (band = 0; band < totbnd; band++)
	    {

/*  Add up all input pixels in blocks corresponding to output pixels  */

	    for (is = 0; is < nsi; is++)
		{
		if ((iline % yfact == 0) && (is % xfact == 0))
		    outbuf[offseto[band] + is / xfact] =
		      inbuf[offseti[band] + is];
		else
		    outbuf[offseto[band] + is / xfact] +=
		      inbuf[offseti[band] + is];
		}
	    }

/*  Compute means for each output block  */

	if (((iline + 1) % yfact == 0) || iline == (nli - 1))
	    {
	    for (band = 0; band < totbnd; band++)
		{
		blksize = xfact * (1 + (iline % yfact));
		for (os = 0; os < nso; os++)
		    {
		    if (os == (nso - 1))
			blksize = (1 + (nsi - 1) % xfact)
			  * (1 + (iline % yfact));
		    outbuf[offseto[band] + os] /= blksize;
		    if (roundflag)
			outbuf[offseto[band] + os]
			  = floor (outbuf[offseto[band] + os] + 0.5);
		    }

/*  Update output min/max values  */

		c_pxmnmx ((long *) (outbuf + offseto[band]),
		  &nso, mmval[0] + band, mmval[1] + band, &dtype);
		}

/*  Write output lines for current input line and band  */

	    if (c_estep (&gdo) != E_SUCC)
		c_clean (vblock, "zoom", hosout, &one);
	    }
	}

    /*  Stop processing messages and close image files  */

    c_stoptm ();
    c_egclse (&gdi);
    c_egclse (&gdo);

/*  Update DDR with new scale and corner coordinates  */

    xratio = xfact;
    yratio = yfact;

    if (c_upddr (vblock, hosin, hosout[0], &nmrimg, window, bands,
      nbands, &yratio, &xratio, &totbnd, &zero) != E_SUCC)
	c_clean (vblock, "zoom", hosout, &one);
    if ((bddr = (struct BDDR *) malloc (totbnd * sizeof (struct BDDR)))
       == NULL)
	c_lasmsg (FALSE, "zoom-alloc",
	  "Error allocating dynamic memory");
    else
	for (band = 1; band <= totbnd; band++)
	    if (c_getbdr (hosout[0], bddr + band - 1, &band) != E_SUCC)
		break;
    if ( (bddr == NULL) || (band <= totbnd)
      || (c_getddr (hosout[0], &ddr) != E_SUCC) )
	c_clean (vblock, "zoom", hosout, &one);
    ddr.upleft[0] -= ((yratio - 1) / yratio) * (ddr.pdist_y / 2.0);
    ddr.upright[0] = ddr.upleft[0];
    ddr.upleft[1] += ((xratio - 1) / xratio) * (ddr.pdist_x / 2.0);
    ddr.loright[1] = ddr.upleft[1];
    ddr.loleft[0] = ddr.loright[0] = ddr.upleft[0]
      - (nlo - 1) * ddr.pdist_y;
    ddr.upright[1] = ddr.loright[1] = ddr.upleft[1]
      + (nso - 1) * ddr.pdist_x;
    if (c_putddr (hosout[0], &ddr) != E_SUCC)
	c_clean (vblock, "zoom", hosout, &one);
    for (band = 0; band < totbnd; band++)
        if (c_putbdr (hosout[0], &bddr[band]) != E_SUCC)
	    c_clean (vblock, "zoom", hosout, &one);

/*  Insert band minima and maxima  */

    if (c_svmnmx (hosout[0], &totbnd, mmval[0], mmval[1], minmax)
      != E_SUCC)
	c_lasmsg (FALSE, "zoom-warn", "Nonfatal error encountered");
    free (inbuf);
    free (outbuf);
    free (mmval[0]);
    free (bddr);

    return E_SUCC;
    }			/* End of Function zmneanred */
