/*  zmrandex -- Expand an image with randomization

NAME:
    zmrandex

PURPOSE:
    Called by ZOOM for subcommand RANDEX to expand an image by pixel
    replication, multiplying the input pixel by a different randomly
    selected value for each output pixel.

ARGUMENTS:
    struct PARBLK
	    *vblock	(in)	TAE parameter block.
    char hosin[][CMLEN]	(in)	Host names of input images.
    char hosout[][CMLEN](in)	Host name of output image.
    long    nmrimg	(in)    Number of input images.
    long    (*window)[4] (in)	Window specification output image.
    long    (*bands)[MAXBND+1]
			(in)	Bands for each input image.
    long    *nbands	(in)	Number of bands per input image.
    long    totbnd	(in)	Total number of input bands.
    long    odtype	(in)	Data type of output image.
    long    xfact	(in)	Expansion factor for the x-axis.
    long    yfact	(in)	Expansion factor for the y-axis.
    long    msgtim	(in)	Processing progress message interval.
    char    *minmax	(in)	Min/max flag.
    char    *randfunc	(in)	Name of distribution function.
    float   coeff[]	(in)	Coefficients for distribution.
    float   range[2]	(in)	Minimum and maximum multipliers.
    char    select[]	(in)	Multiplier selection option.
    long    nbin	(in)	Number of bins in multiplier table.
    long    seed	(in)	Seed for random number generator.

RETURN VALUE:
    E_SUCC	Normal completion.
		(Function aborts on fatal error).

FILE REFERENCES:
    hosin	Input image file(s).
    hosout	Output image file.

PROGRAM HISTORY:                                  
    1998 May	Initial version, R. A. White, PSU/ESSC.
    1999 Apr	Changed c_errmsg() calls to c_lasmsg(), round fractions
		>= 0.5 upwards, R. A. White.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.5. 

PROJECT:
    LAS at PSU/ESSC

INFORMATIONAL AND ERROR MESSAGES:

    These are described in zoom.pdf.

EXTERNAL FUNCTIONS CALLED:

    zmbldtbl	Build table of multipliers distributed as specified by
		arguments randfunc, coeff, and range.

ALGORITHM DESCRIPTION:
    Create table of multipliers <zmbldtbl>
    Open input image(s) for sequential access
    Allocate input buffer
    Define input group
    Allocate buffer to hold yfact lines of output
    Initialize random number generator
    Initialize output min/max
    Open output for random access
    Start processing messages
    For each input line
        For each input band
	    For each input pixel
	        If all multipliers are to be used (SELECT = "ALL")
		    Randomly assign multipliers to output pixels
		Else
		    Randomly select multipliers from table
		    If renormalization requested (SELECT = "RENORM")
			Sum the multipliers and compute the average
			Divide each multiplier by the average
		For each output pixel from current input pixel
		    Compute output pixel as input pixel times the
		      multiplier selected for output pixel
		    If floating point value is to be rounded to integer
		        Round fraction >= 0.5 upwards
	    For each output line corresponding to input line
		Update output min/max
		Write output line
    Stop processing messages
    Close image files
    Update DDR with new scale and corner coordinates
    Insert band minima and maxima
    									
***************   End of Prolog for zmrandex   ***************	*/
									/*
Header Files:			       					*/

#include	<ctype.h>	/* Character classification functions	*/
#include	<math.h>	/* Mathematical functions		*/
#include	<stdio.h>	/* Standard I/O definitions		*/
#include	<stdlib.h>	/* Standard C functions			*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"util.h"	/* LAS utility functions		*/
#include	"pixman.h"	/* LAS pixel-manipulation functions	*/
						   			/*
Local Program Constants and Macros:					*/
#define  ABORT \
	   c_lasmsg (TRUE, "zoom-fatal", "Fatal error encountered")
									/*
External Function Prototypes Not Included in Header Files:		*/
     void	zmbldtbl (char *randfunc, float coeff[], float range[],
			  long nbin, float multtab[]);
									/*
External Variable Declarations:						*/
    extern long	curent;		/* current processing line */
    extern long curcat;		/* current processing band */
									/*
***************   Start of Function zmrandex ()   ***************	*/

long zmrandex (
    struct PARBLK *vblock,	/* TAE parameter block */
    char hosin[][CMLEN],	/* Host names of input images */
    char hosout[][CMLEN],	/* Host name of output image */
    long	nmrimg,		/* Number of input images */
    long	(*window)[4],	/* Window specification output image */
    long	(*bands)[MAXBND+1], /* Bands for each input image */
    long	*nbands,	/* Number of bands per input image */
    long	totbnd,		/* Total number of input bands */
    long	odtype,		/* Data type of output image */
    long	xfact,		/* Expansion factor for the x-axis */
    long	yfact,		/* Expansion factor for the y-axis */
    long	msgtim,		/* Processing progress message interval */
    char	*minmax,	/* Min/max flag */
    char	*randfunc,	/* Name of distribution function */
    float	coeff[],	/* Coefficients for distribution */
    float	range[2],	/* Minimum and maximum multipliers */
    char	select[],	/* Multiplier selection option */
    long	nbin,		/* Number of bins in multiplier table */
    long	seed)		/* Seed for random number generator */
    {
	                                              			/*
Local Variable Definitions:						*/

    long	access;		/* Code for image access type */
    long	band;		/* Sequential input band number */
    long	blksize;	/* Size of output block for input pixel */
    long	bufsize;	/* Size of input buffer */
    long	dtype;		/* Code for internal data type */
    struct FDESC *fdi[MAXIMG+1]; /* File descriptor(s) for input image */
    struct FDESC *fdo;		/* File descriptor for output image */
    struct GDESC *gdi;		/* Group descriptor for input image(s) */
    long	i;		/* Loop counter/index */
    long	image;		/* Current input image */
    unsigned char *inbuf;	/* Buffer for input image data */
    long	is;		/* Input sample number */
    long	iline;		/* Current input line */
    float	ival;		/* Value of input pixel */
    long	j;		/* Temporary storage for weight index */
    double	*mmval[2];	/* Arrays for output minima and maxima */
    long	msgformat;	/* Processing progress message format */
    float	*multtbl;	/* Table of multipliers */
    long	nli;		/* Number of lines in input window */
    long	nlo;		/* Number of output lines */
    long	nsi;		/* Number of samples per input line */
    long	nso;		/* Number of samples per output line */
    long	oband;		/* Output band, counting from 1 */
    long	offset[MAXBND];	/* Offsets into input buffer */
    long	oline;		/* First line of current output block */
    long	one = 1;	/* Constant one */
    long	opt[2];		/* Options for c_eopen () functions */
    long	obufsize;	/* Size of output buffer, in pixels */
    float	*outbuf;	/* Buffer for output data */
    float	oval;		/* Value of output pixel */
    long	roundflag;	/* Flag for integer output data type */
    float	tmp;		/* Temporary storage for multiplier */
    float	*wghts;		/* Array of weights for output block */
    float	wghtave;	/* Average of weights for block */
    double	xratio;		/* Ratio of output to input x */
    double	yratio;		/*  and y pixel dimensions    */
    long	zero = 0;	/* Constant zero */

/*  Create table of multipliers  */

    select[0] = toupper (select[0]);
    if ((select[1] = toupper (select[1])) == 'E')
	select[0] = 'Y';
    blksize = yfact * xfact;
    if (select[0] == 'A')
	nbin = blksize;
    if ((multtbl = (float *) malloc ((nbin + blksize) * sizeof (float)))
	== NULL)
	c_lasmsg (TRUE, "zoom-alloc", "Error allocating buffer space");
    wghts = multtbl + nbin;
    zmbldtbl (randfunc, coeff, range, nbin, multtbl);

/*  Open input image(s) for sequential access  */

    access = IREAD;
    opt[0] = opt[1] = 0;
    nli = window[0][2];
    nsi = window[0][3];
    dtype = EREAL;

    for (band = 0, image = 0; image < nmrimg; image++)
	{
	if (c_eopens (&fdi[image], hosin[image], bands[image],
	  &window[image][0], &window[image][1] , &nli ,&nsi,
	  &offset[band], &access, &dtype, &zero, opt) != E_SUCC)
	    ABORT;
	band += nbands[image];
	}
    fdi[nmrimg] = 0;

/*  Allocate input buffer  */

    if ((inbuf = (unsigned char *) c_ealloc (fdi, &bufsize)) == NULL)
	c_lasmsg (TRUE, "zoom-alloc",
	  "Error allocating dynamic memory");

/*  Define input group  */

    if (c_egroup (fdi, &gdi, inbuf, &bufsize) != E_SUCC)
	ABORT;

/*  Allocate buffer for output  */

    obufsize = nsi * yfact * xfact;
    if ((outbuf = (float *) malloc (obufsize * sizeof (*outbuf)))
      == NULL) 
	c_lasmsg (TRUE, "zoom-alloc",
	  "Error allocating dynamic memory");

/*  Initialize random number generator  */

    srand48 (seed);

/*  Initialize output min/max array  */

    if ((mmval[0] = (double *) malloc (2 * totbnd * sizeof (double)))
      == NULL)
	c_lasmsg (TRUE, "zoom-alloc",
	  "Error allocating dynamic memory");
    mmval[1] = mmval[0] + totbnd;
    if (c_inmnmx (&odtype, &totbnd, mmval[0], mmval[1]) != E_SUCC)
	ABORT;

/*  Open output image for random access  */

    nlo = yfact * nli;
    nso = xfact * nsi;
    access = IWRITE;
    opt[0] = odtype;
    if (c_eopenr (&fdo, hosout[0], &totbnd, &one, &one, &nlo, &nso,
      &access, &dtype, opt, &yfact) != E_SUCC)
	ABORT;

/*  Start processing messages  */

    if (msgtim > 0)
	{
	if (totbnd > 1)
	    {
	    msgformat = ONECAT;
	    c_inittm ("ZOOM", "processing", "line", "band", "", &nli,
		    &totbnd, &msgtim, &msgformat);
	    }
	else
	    {
	    msgformat = GEN2;
	    c_inittm ("ZOOM", "processing", "line", "band", "", &nli,
		    &totbnd, &msgtim, &msgformat);
	    }
	curent = curcat = 1;
	c_startm ();
	}

/*  Process successive bands for one line at a time  */

    roundflag = ((odtype == EREAL) || (odtype == EDOUBLE)) ? 0 : 1;
    for (iline = 1; iline <= nli; iline++, curent++)
	{
	oline = (iline - 1) * yfact + 1;
	if (c_estep (&gdi) != E_SUCC)
	    {
	    c_eclose(&fdo);
	    c_clean (vblock, "zoom", hosout, &one);
	    }
	for (band = 0, curcat = 1; band < totbnd; band++, curcat++)
	    {
	    for (is = 0; is < nsi; is++)
		{

/*  Select and normalize multipliers for input pixel  */

		if (select[0] == 'A')
		    {
		    for (i = 0; i < blksize; i++)
			wghts[i] = multtbl[i];
		    for (i = blksize; i > 1; )
			{
			j = lrand48() % (i--);
			tmp = wghts[i];
			wghts[i] = wghts[j];
			wghts[j] = tmp;
			}
		    }
		else
		    {
		    for (i = 0, wghtave = 0.0; i < blksize; i++)
			{
			wghts[i] = multtbl[lrand48() % nbin];
			wghtave += wghts[i];
			}
		    if (select[0] == 'Y')
			{
			wghtave /= blksize;
			for (i = 0; i < blksize; i++)
			    wghts[i] /= wghtave;
			}
		    }

/*  Use multipliers to compute output values for each input pixel  */

		ival = ((float *) inbuf)[offset[band] + is];
		for (i = 0; i < blksize; i++)
		    {
		    oval = ival * wghts[i];
		    if (roundflag)
			oval = floor (oval + 0.5);
		    outbuf[is * xfact + (nso - xfact) * (i / xfact) + i]
		      = oval;
		    }
		}

/*  Update output min/max values  */

	    c_pxmnmx ((long *) outbuf, &obufsize,
	      mmval[0] + band, mmval[1] + band, &dtype);

/*  Write output lines for current input line and band  */

	    oband = band + 1;
	    if (c_ewrite (&fdo, &oband, &oline,
	      (unsigned char *) outbuf, &yfact) != nso * yfact)
		{
		c_eclose(&fdo);
		c_clean (vblock, "zoom", hosout, &one);
		}
	    }
	}

    /*  Stop processing messages and close image files  */

    c_stoptm ();
    c_egclse (&gdi);
    c_eclose (&fdo);

/*  Update DDR with new scale and corner coordinates  */

    xratio = 1.0 / xfact;
    yratio = 1.0 / yfact;
    if (c_upddr (vblock, hosin, hosout[0], &nmrimg, window, bands,
      nbands, &yratio, &xratio, &totbnd, &zero) != E_SUCC)
	c_clean(vblock,"zoom",hosout,&one);

/*  Insert band minima and maxima  */

    if (c_svmnmx (hosout[0], &totbnd, mmval[0], mmval[1], minmax)
      != E_SUCC)
	c_lasmsg (FALSE, "zoom-warn", "Nonfatal error encountered");
    free (inbuf);
    free (outbuf);
    free (multtbl);
    free (mmval[0]);

    return E_SUCC;
    }			/* End of Function zmrandex */
